﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using Microsoft.Research.Kinect.Nui;
using Coding4Fun.Kinect.Wpf;
using AForge;
using AForge.Math.Geometry;
using AForge.Vision.GlyphRecognition;

namespace DepthTracker
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        private enum ScanMode
        {
            None,
            Obstacle,
            Corridor
        }

        private Runtime m_NUI;
        private bool m_UsingZoom;
        private ScanMode m_Mode = ScanMode.Corridor;

        private int totalFrames = 0;
        private int lastFrames = 0;
        private DateTime lastTime = DateTime.MaxValue;

        private ObstacleDetector m_ObstacleDetector;
        private CorridorFinder m_CorridorFinder;
        private GlyphDetector m_GlyphDetector;

        private Byte[] m_ColorFrame;

        const float MinDepthDistance = 800;
        const float MaxDepthDistance = 4000;
        const float MaxDepthDistanceOffset = MaxDepthDistance - MinDepthDistance;

        public MainWindow()
        {
            InitializeComponent();

            try
            {
                Rect bounds = Properties.Settings.Default.WinPosition;
                this.Top = bounds.Top;
                this.Left = bounds.Left;
                this.Height = bounds.Height;
                this.Width = bounds.Width;
            }
            catch
            {
            }
        }

        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            try
            {
                m_NUI = Runtime.Kinects[0];
                m_NUI.Initialize(RuntimeOptions.UseDepth | RuntimeOptions.UseColor);
            }
            catch (InvalidOperationException)
            {
                System.Windows.MessageBox.Show("Runtime initialization failed. Please make sure Kinect device is plugged in.");
                return;
            }


            try
            {
                m_NUI.VideoStream.Open(ImageStreamType.Video, 2, ImageResolution.Resolution640x480, ImageType.Color);
                m_NUI.DepthStream.Open(ImageStreamType.Depth, 2, ImageResolution.Resolution80x60, ImageType.Depth);
            }
            catch (InvalidOperationException)
            {
                System.Windows.MessageBox.Show("Failed to open stream. Please make sure to specify a supported image type and resolution.");
                return;
            }

            rbCorridor.IsChecked = true;

            lastTime = DateTime.Now;

            m_NUI.DepthFrameReady += new EventHandler<ImageFrameReadyEventArgs>(nui_DepthFrameReady);
            m_NUI.VideoFrameReady += new EventHandler<ImageFrameReadyEventArgs>(nui_ColorFrameReady);

            txtStatus.Text = m_NUI.Status.ToString();
            slTilt.Value = Properties.Settings.Default.CameraAngle;
            m_NUI.NuiCamera.ElevationAngle = (int)slTilt.Value;
        }

        private void Window_Closing(object sender, System.ComponentModel.CancelEventArgs e)
        {
            Properties.Settings.Default.WinPosition = this.RestoreBounds;
            Properties.Settings.Default.Save();
        }

        private void btnClose_Click(object sender, RoutedEventArgs e)
        {
            this.Close();
        }

        private void btnTilt_Click(object sender, RoutedEventArgs e)
        {
            m_NUI.NuiCamera.ElevationAngle = (int)slTilt.Value;
            Properties.Settings.Default.CameraAngle = slTilt.Value;
            Properties.Settings.Default.Save();
        }

        private void cbUsingZoom_Click(object sender, RoutedEventArgs e)
        {
            m_UsingZoom = (bool)cbUsingZoom.IsChecked;
        }

        private void rbCorridor_Checked(object sender, RoutedEventArgs e)
        {
            SetupCorridorFinder();
        }

        private void rbObstacle_Checked(object sender, RoutedEventArgs e)
        {
            SetupObstacleDetector();
            m_Mode = ScanMode.Obstacle;
        }

        private void nui_ColorFrameReady(object sender, ImageFrameReadyEventArgs e)
        {
            // 32-bit per pixel, RGBA image
            imgVideo.Source = e.ImageFrame.ToBitmapSource();

            List<HotZoneSpec> hotzones = m_GlyphDetector.Process(e.ImageFrame);
            ClearCanvas(cvsVideo);
            foreach (HotZoneSpec hz in hotzones)
            {
                DrawHotZone(cvsVideo, .5, Brushes.Yellow, hz);
                
                // Get glyph's bearing
                double X = hz.Rectangle.X + (hz.Rectangle.Width / 2);
                double bearing = Math.Round(30 * ((X - 320.0) / 320.0), 0);
                DrawBearing(cvsVideo, .5, Brushes.Lime, X, bearing.ToString());
            }

            if (tabControl.SelectedIndex == 1)
            {
                // Calculate and display frames per second
                DisplayFrameRate();
            }
        }

        private void nui_DepthFrameReady(object sender, ImageFrameReadyEventArgs e)
        {
            System.Diagnostics.Debug.WriteLine(System.Threading.Thread.CurrentThread.Name);

            // Build new ColorFrame that contains color information for all pixels in image
            // Height x Width x 4 (Red, Green, Blue, empty byte)
            m_ColorFrame = new byte[e.ImageFrame.Image.Height * e.ImageFrame.Image.Width * 4];

            switch (m_Mode)
            {
                case ScanMode.None:
                    return;

                case ScanMode.Obstacle:
                    // Process depth data through Obstacle Detector
                    List<HotZoneSpec> hits = m_ObstacleDetector.Process(e.ImageFrame);
                    break;

                case ScanMode.Corridor:
                    double x = m_CorridorFinder.Process(e.ImageFrame);
                    if (x == -1)
                        ClearCanvas(cvsDepth);
                    else
                        DrawCorridor(cvsDepth, cvsDepth.Width / 80, Brushes.White, x);
                    break;
            }

            // Display colored depth image
            PlanarImage image = e.ImageFrame.Image;
            imgDepth.Source = BitmapSource.Create(image.Width, image.Height, 96, 96, PixelFormats.Bgr32, null,
                m_ColorFrame, image.Width * PixelFormats.Bgr32.BitsPerPixel / 8);

            if (tabControl.SelectedIndex == 0)
            {
                // Calculate and display frames per second
                DisplayFrameRate();
            }
        }

        private void ObstacleDetectorCallback(ObstacleDetector.ScanResult result)
        {
            // Hardcoded locations to Blue, Green, Red (BGR) index positions       
            const int BlueIndex = 0;
            const int GreenIndex = 1;
            const int RedIndex = 2;

            var heightOffset = result.Y * result.ImageFrameWidth;
            var index = (int)((result.X + heightOffset) * 4);

            // Set pixel color based on distance or in a hot zone
            if (result.Distance == 0)
            {
                // Unknown distance, show in black
                m_ColorFrame[index + BlueIndex] = 64;
                m_ColorFrame[index + GreenIndex] = 0;
                m_ColorFrame[index + RedIndex] = 0;
            }
            else if(result.Hit)
            {
                // In hot zone, show in red
                m_ColorFrame[index + BlueIndex] = 0;
                m_ColorFrame[index + GreenIndex] = 0;
                m_ColorFrame[index + RedIndex] = 255;
            }
            else
            {
                // Use gray scale coloring for monochromatic histogram
                var intensity = CalculateIntensityFromDepth(result.Distance);
                m_ColorFrame[index + BlueIndex] = intensity;
                m_ColorFrame[index + GreenIndex] = intensity;
                m_ColorFrame[index + RedIndex] = intensity;
            }
        }

        private int GetDistance(byte firstFrame, byte secondFrame)
        {
            int distance = (int)firstFrame | secondFrame << 8;
            return distance;
        }

        public byte CalculateIntensityFromDepth(double distance)
        {
            // formula for calculating monochrome intensity for histogram
            if (distance == 0)
                return 0;
            else
                return (byte)(255 - (255 * Math.Max(distance - MinDepthDistance, 0) / (MaxDepthDistanceOffset)));
        }

        private void SetupObstacleDetector()
        {
            m_Mode = ScanMode.Obstacle;

            m_ObstacleDetector = new ObstacleDetector();
            m_ObstacleDetector.Scan += ObstacleDetectorCallback;

            m_ObstacleDetector.AddHotZone("Left", 0, 25, 15, 20, 900, 1400);
            m_ObstacleDetector.AddHotZone("CenterLeft", 15, 25, 25, 25, 800, 1000);
            m_ObstacleDetector.AddHotZone("CenterRight", 40, 25, 25, 25, 800, 1000);
            m_ObstacleDetector.AddHotZone("Right", 65, 25, 15, 20, 900, 1400);

            // Display hot zones
            cvsDepth.Children.Clear();
            foreach (HotZoneSpec hz in m_ObstacleDetector.HotZones)
                DrawHotZone(cvsDepth, cvsDepth.Width / 80, Brushes.White, hz);
        }

        private void SetupCorridorFinder()
        {
            m_Mode = ScanMode.Corridor;

            double max = 850;
            if ((bool)cbUsingZoom.IsChecked)
                max = 1200;
            m_CorridorFinder = new CorridorFinder(20, 30, 40, 800, max);
            m_CorridorFinder.Scan += ObstacleDetectorCallback;

            SetupGlyphDetector();
        }

        private void SetupGlyphDetector()
        {
            m_GlyphDetector = new GlyphDetector(5);
            m_GlyphDetector.AddGlyph("U", "0000001110010000111000000");
            m_GlyphDetector.AddGlyph("T", "0000001110001000010000000");
        }

        public void DrawHotZone(Canvas canvas, double scale, Brush stroke, HotZoneSpec hz)
        {
            System.Windows.Shapes.Rectangle rect = new System.Windows.Shapes.Rectangle();
            rect.Stroke = stroke;

            rect.Width = hz.Rectangle.Width * scale;
            rect.Height = hz.Rectangle.Height * scale;
            Canvas.SetLeft(rect, hz.Rectangle.Left * scale);
            Canvas.SetTop(rect, hz.Rectangle.Top * scale);
            canvas.Children.Add(rect);

            TextBlock tb = new TextBlock();
            tb.Text = hz.Name;
            tb.Foreground = stroke;
            tb.Width = rect.Width;
            tb.Height = 44;
            tb.TextWrapping = TextWrapping.Wrap;
            tb.TextAlignment = TextAlignment.Center;
            Canvas.SetLeft(tb, hz.Rectangle.Left * scale);
            Canvas.SetTop(tb, (hz.Rectangle.Top * scale) + ((rect.Height - tb.Height) / 2));
            canvas.Children.Add(tb);
        }

        public void DrawBearing(Canvas canvas, double scale, Brush stroke, double X, string text)
        {
            System.Windows.Shapes.Line line = new System.Windows.Shapes.Line();
            line.Stroke = stroke;

            line.X1 = 1;
            line.X2 = 1;
            line.Y1 = 1;
            line.Y2 = 240;
            line.HorizontalAlignment = HorizontalAlignment.Left;
            line.VerticalAlignment = VerticalAlignment.Center;
            line.StrokeThickness = 1;
            Canvas.SetLeft(line, X * scale);
            Canvas.SetTop(line, 0);
            canvas.Children.Add(line);

            TextBlock tb = new TextBlock();
            tb.Text = text;
            tb.Foreground = stroke;
            tb.Width = 30;
            tb.Height = 44;
            tb.TextWrapping = TextWrapping.Wrap;
            tb.TextAlignment = TextAlignment.Center;
            Canvas.SetLeft(tb, X * scale);
            Canvas.SetTop(tb, 220);
            canvas.Children.Add(tb);

        }

        private void DrawCorridor(Canvas canvas, double scale, Brush stroke, double X)
        {
            canvas.Children.Clear();

            System.Windows.Shapes.Rectangle rect = new System.Windows.Shapes.Rectangle();
            rect.Stroke = stroke;

            rect.Width = m_CorridorFinder.Corridor.Rectangle.Width * scale;
            rect.Height = m_CorridorFinder.Corridor.Rectangle.Height * scale;

            double offset = m_CorridorFinder.Corridor.Rectangle.Width / 2.0;
            Canvas.SetLeft(rect, (X - offset) * scale);
            Canvas.SetTop(rect, m_CorridorFinder.Corridor.Rectangle.Y * scale);
            canvas.Children.Add(rect);

            double bearing = 30 * ((X - 40.0) / 40.0);

            TextBlock tb = new TextBlock();
            tb.Text = bearing.ToString();
            tb.Foreground = stroke;
            tb.Width = rect.Width;
            tb.Height = 44;
            tb.TextWrapping = TextWrapping.Wrap;
            tb.TextAlignment = TextAlignment.Center;
            Canvas.SetLeft(tb, (X - offset) * scale);
            Canvas.SetTop(tb, (m_CorridorFinder.Corridor.Rectangle.Y * scale) + ((rect.Height - tb.Height) / 2));
            canvas.Children.Add(tb);
        }

        private void ClearCanvas(Canvas canvas)
        {
            canvas.Children.Clear();
        }

        private void DisplayFrameRate()
        {
            // Calculate and display frames per second
            ++totalFrames;
            DateTime cur = DateTime.Now;
            if (cur.Subtract(lastTime) > TimeSpan.FromSeconds(1))
            {
                int frameDiff = totalFrames - lastFrames;
                lastFrames = totalFrames;
                lastTime = cur;
                txtFrameRate.Text = frameDiff.ToString() + " fps";
            }
        }
    }
}
