﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO.Ports;

namespace KBotController
{
    public class Controller : IDisposable
    {
        private SerialPort m_Port;
        private bool m_Disposed = false;

        public enum Direction
        {
            Halt,
            Forward,
            Reverse,
            SpinCW,
            SpinCCW
        }

        // Status
        private int m_KinectPos;
        private bool m_DrivePower;
        private int m_Speed;
        private Direction m_Direction;
        private int m_WheelPosLeft;
        private int m_WheelPosRight;
        private bool m_HasArrived;

        // Sensors
        private int m_IRLeft;
        private int m_IRRight;
        private int m_PingLeft;
        private int m_PingCenter;
        private int m_PingRight;

        public Controller(string portName)
        {
            SetComPort(portName);
        }

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        private void Dispose(bool disposing)
        {
            // Check to see if Dispose has already been called.
            if (!this.m_Disposed)
            {
                // If disposing equals true, dispose all managed 
                // and unmanaged resources.
                if (disposing)
                {
                    // Dispose managed resources.
                    m_Port.Close();
                }

                // Call the appropriate methods to clean up 
                // unmanaged resources here.
                // If disposing is false, 
                // only the following code is executed.
            }
            m_Disposed = true;
        }
        
        public void SetPower(bool IsOn)
        {
            if(IsOn)
                SendCommand("P1");
            else
                SendCommand("p1");
        }

        public void SetSpeed(int speed)
        {
            SendCommand("S" + speed.ToString());
        }

        public void MoveForward()
        {
            SendCommand("FW");
        }

        public void MoveForward(int distance)
        {
            SendCommand("FX" + distance.ToString());
        }

        public void MoveForward(int speed, int distance)
        {
            SetSpeed(speed);
            MoveForward(distance);
        }

        public void MoveReverse()
        {
            SendCommand("BK");
        }

        public void MoveReverse(int speed)
        {
            SetSpeed(speed);
            MoveReverse();
        }

        public void Halt()
        {
            SendCommand("HL");
        }

        public void EmergencyHalt()
        {
            SendCommand("EH");
        }

        public void TurnLeft(int bias)
        {
            SendCommand("LX" + bias.ToString());
        }

        public void TurnRight(int bias)
        {
            SendCommand("RX" + bias.ToString());
        }

        public void SetKinectPosition(int pos)
        {
            SendCommand("KX" + pos.ToString());
        }

        public bool UpdateSensorData()
        {
            List<string> dataItems = GetTelemetry("1");
            if (dataItems != null && dataItems.Count == 5)
            {
                m_IRLeft = int.Parse(dataItems[0]);
                m_IRRight = int.Parse(dataItems[1]);
                m_PingLeft = int.Parse(dataItems[2]);
                m_PingCenter = int.Parse(dataItems[3]);
                m_PingRight = int.Parse(dataItems[4]);
                return true;
            }
            return false;
        }

        public bool UpdateStatus()
        {
            List<string> dataItems = GetTelemetry("2");
            if (dataItems != null && dataItems.Count == 7)
            {
                m_KinectPos = int.Parse(dataItems[0]);
                m_DrivePower = (int.Parse(dataItems[1]) == 1) ? true : false;
                m_Speed = int.Parse(dataItems[2]);
                m_Direction = (Direction)int.Parse(dataItems[3]);
                m_WheelPosLeft = int.Parse(dataItems[4]);
                m_WheelPosRight = int.Parse(dataItems[5]);
                m_HasArrived = (int.Parse(dataItems[6]) == 1) ? true : false;
                return true;
            }
            return false;
        }

        #region Properties

        public int IRLeft
        {
            get { return m_IRLeft; }
        }

        public int IRRight
        {
            get { return m_IRRight; }
        }

        public int PingLeft
        {
            get { return m_PingLeft; }
        }

        public int PingCenter
        {
            get { return m_PingCenter; }
        }

        public int PingRight
        {
            get { return m_PingRight; }
        }

        public int KinectPosition
        {
            get { return m_KinectPos; }
        }

        public bool DrivePower
        {
            get { return m_DrivePower; }
        }

        public int CurrentSpeed
        {
            get { return m_Speed; }
        }

        public Direction CurrentDirection
        {
            get { return m_Direction; }
        }

        public int WheelPositionLeft
        {
            get { return m_WheelPosLeft; }
        }

        public int WheelPositionRight
        {
            get { return m_WheelPosRight; }
        }

        public bool HasArrived
        {
            get { return m_HasArrived; }
        }

        #endregion

        #region Helpers

        private void SetComPort(string portName)
        {
            if (m_Port != null)
                m_Port.Close();

            m_Port = new SerialPort(portName, 19200);
            m_Port.DataBits = 8;
            m_Port.StopBits = StopBits.One;
            m_Port.Parity = Parity.None;
            m_Port.Handshake = Handshake.None;
            m_Port.WriteTimeout = 500;
            m_Port.ReadTimeout = 1000;
            m_Port.NewLine = "\r";

            m_Port.Open();
        }

        private void SendCommand(string cmd)
        {
            if (m_Port.IsOpen)
            {
                string msg = ">RK" + cmd;
                m_Port.WriteLine(msg);
            }
        }

        private List<string> GetTelemetry(string frame)
        {
            SendCommand("T" + frame);

            string tel = "";
            try
            {
                tel = m_Port.ReadLine();
            }
            catch
            {
            }

            if (tel.StartsWith("<RK"))
            {
                string[] data = tel.Substring(3).Split(',');
                return data.ToList<string>();
            }

            return null;
        }

        #endregion
    }
}
