﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using Microsoft.Research.Kinect.Nui;
using Microsoft.Research.Kinect.Audio;
using Microsoft.Speech.AudioFormat;
using Microsoft.Speech.Recognition;
using Coding4Fun.Kinect.Wpf;

namespace KBotTracker
{
    class Cortex
    {
        public enum TrackingMode
        {
            Visual,
            Audio,
            None
        }

        public enum TrackStatus
        {
            Tracking,
            NoFrame,
            NoPerson,
            NoJoint
        }

        public struct TrackingResult
        {
            public TrackingMode ResultType { get; set; }
            public TrackStatus Status { get; set; }
            public int Bearing { get; set; }
            public double Distance { get; set; }
            public SkeletonQuality Quality { get; set; }
            public int TrackingID { get; set; }
        }

        private Runtime m_NUI;
        private Controller m_KBot;

        private TrackingMode m_TrackingMode = TrackingMode.None;
        private bool m_ActiveTracking = false;

        private Task m_TrackingTask;
        private CancellationTokenSource m_TrackingTaskCancel;

        private Thread m_AudioTrackThread;
        private bool m_StopAudioTracking = false;
        private const string RecognizerId = "SR_MS_en-US_Kinect_10.0";

        private string m_PortName;
        private int m_CameraAngle;
        private bool m_UsingZoom = false;

        private int m_CurrentKinectPos;
        private TrackingResult m_LastVisualTrack;
        private TrackingResult m_LastAudioTrack;

        private bool m_Disposed = false;

        public delegate void TrackingHandler(TrackingResult result);
        public event TrackingHandler VisualTrack;
        public event TrackingHandler AudioTrack;

        public Cortex(string portName, int cameraAngle)
        {
            m_PortName = portName;
            m_CameraAngle = cameraAngle;

            // Initialize the KBot Propeller controller
            m_KBot = new Controller(m_PortName);
            SetKinectPosition(90);

            // Start the person tracking task
            m_TrackingTaskCancel = new CancellationTokenSource();
            m_TrackingTask = new Task(new Action(VisualTracking), m_TrackingTaskCancel.Token);
            m_TrackingTask.Start();

            // Start audio tracking thread
            m_AudioTrackThread = new Thread(new ThreadStart(AudioTracking));
            m_AudioTrackThread.SetApartmentState(ApartmentState.MTA);
            m_AudioTrackThread.Start();
        }

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        private void Dispose(bool disposing)
        {
            // Check to see if Dispose has already been called.
            if (!this.m_Disposed)
            {
                // If disposing equals true, dispose all managed 
                // and unmanaged resources.
                if (disposing)
                {
                    m_TrackingTaskCancel.Cancel();

                    m_StopAudioTracking = true;
                    m_AudioTrackThread.Join();

                    // Dispose managed resources.
                    m_KBot.Dispose();
                    m_NUI.Uninitialize();
                }

                // Call the appropriate methods to clean up 
                // unmanaged resources here.
                // If disposing is false, 
                // only the following code is executed.
            }
            m_Disposed = true;
        }

        public void SetSensorScan(bool IsOn)
        {
            m_KBot.SetSensorScan(IsOn);
        }

        public void SetPower(bool IsOn)
        {
            m_KBot.SetPower(IsOn);
        }

        public void SetSpeed(int speed)
        {
            m_KBot.SetSpeed(speed);
        }

        public void MoveForward()
        {
            m_KBot.MoveForward();
        }

        public void MoveForward(int distance)
        {
            m_KBot.MoveForward(distance);
        }

        public void MoveForward(int speed, int distance)
        {
            m_KBot.MoveForward(speed, distance);
        }

        public void MoveReverse()
        {
            m_KBot.MoveReverse();
        }

        public void MoveReverse(int speed)
        {
            m_KBot.MoveReverse(speed);
        }

        public void Halt()
        {
            m_KBot.Halt();
        }

        public void EmergencyHalt()
        {
            m_KBot.EmergencyHalt();
        }

        public void TurnLeft(int bias)
        {
            m_KBot.TurnLeft(bias);
        }

        public void TurnRight(int bias)
        {
            m_KBot.TurnRight(bias);
        }

        public void SetCameraAngle(int tilt)
        {
            m_NUI.NuiCamera.ElevationAngle = tilt;
        }

        #region Properties

        public TrackingResult LastVisualTrack
        {
            get { return m_LastVisualTrack; }
        }

        public TrackingResult LastAudioTrack
        {
            get { return m_LastAudioTrack; }
        }

        public int IRLeft
        {
            get { return m_KBot.IRLeft; }
        }

        public int IRRight
        {
            get { return m_KBot.IRRight; }
        }

        public int PingLeft
        {
            get { return m_KBot.PingLeft; }
        }

        public int PingCenter
        {
            get { return m_KBot.PingCenter; }
        }

        public int PingRight
        {
            get { return m_KBot.PingRight; }
        }

        public bool DrivePower
        {
            get { return m_KBot.DrivePower; }
        }

        public int CurrentSpeed
        {
            get { return m_KBot.CurrentSpeed; }
        }

        public Controller.Direction CurrentDirection
        {
            get { return m_KBot.CurrentDirection; }
        }

        public int WheelPositionLeft
        {
            get { return m_KBot.WheelPositionLeft; }
        }

        public int WheelPositionRight
        {
            get { return m_KBot.WheelPositionRight; }
        }

        public bool HasArrived
        {
            get { return m_KBot.HasArrived; }
        }

        public TrackingMode CurrentTrackingMode
        {
            get { return m_TrackingMode; }
            set { m_TrackingMode = value; }
        }

        public bool ActiveTracking
        {
            get { return m_ActiveTracking; }
            set
            {
                m_ActiveTracking = value;
                if(!m_ActiveTracking)
                    SetKinectPosition(90);
            }
        }

        public bool UsingZoom
        {
            get { return m_UsingZoom; }
            set { m_UsingZoom = value; }
        }

        public int CurrentServoPos
        {
            get { return m_CurrentKinectPos; }
        }

        #endregion

        #region Helpers

        private void InitializeNui()
        {
            // NUI runtime
            m_NUI = Runtime.Kinects[0];

            // Initialize skeletal tracking
            m_NUI.Initialize(RuntimeOptions.UseSkeletalTracking);

            // Must set to true and set after call to Initialize
            m_NUI.SkeletonEngine.TransformSmooth = true;

            // Use to transform and reduce jitter
            var parameters = new TransformSmoothParameters
            {
                Smoothing = 0.75f,
                Correction = 0.0f,
                Prediction = 0.0f,
                JitterRadius = 0.05f,
                MaxDeviationRadius = 0.04f
            };
            m_NUI.SkeletonEngine.SmoothParameters = parameters;
            m_NUI.SkeletonEngine.IsEnabled = true;
        }

        private void SetKinectPosition(int pos)
        {
            int newPos = pos;
            int delay = Math.Abs((m_CurrentKinectPos - pos)) * 3;

            m_CurrentKinectPos = newPos;
            int inServoRange = Math.Max(0, Math.Min(newPos, 180));
            m_KBot.SetKinectPosition(inServoRange);

            // Wait for servo to move
            Thread.Sleep(delay);
        }

        private void VisualTracking()
        {
            // Initialize the Kinect runtime
            InitializeNui();
            m_NUI.NuiCamera.ElevationAngle = m_CameraAngle;

            while (true)
            {
                // Check for task cancellation
                if(m_TrackingTaskCancel.Token.IsCancellationRequested)
                    break;

                // Update sensor data and robot state
                m_KBot.UpdateSensorData();
                m_KBot.UpdateStatus();

                if (m_TrackingMode == TrackingMode.Visual)
                {
                    // Get the next frame and look for skeletons (in the closet)
                    SkeletonFrame allSkeletons = m_NUI.SkeletonEngine.GetNextFrame(1000);
                    if (allSkeletons != null)
                    {
                        SkeletonData skeleton = (from s in allSkeletons.Skeletons
                                                 where s.TrackingState == SkeletonTrackingState.Tracked
                                                 select s).FirstOrDefault();

                        if (skeleton != null)
                        {
                            var joint = skeleton.Joints[JointID.ShoulderCenter];
                            if (joint.TrackingState == JointTrackingState.Tracked)
                            {
                                // Compute the offset to target from -30 to 30 degrees (right to left)
                                var scaledJoint = joint.ScaleTo(60, 60, .5f, .5f);
                                int offset = Convert.ToInt32(30 - Math.Round(scaledJoint.Position.X, 0));

                                // Compute bearing to target
                                int bearing = m_CurrentKinectPos + offset;

                                // Reposition the servo if target is nearing the visual limits of the Kinect's FOV
                                if (m_TrackingMode == TrackingMode.Visual && m_ActiveTracking && Math.Abs(offset) > 20)
                                {
                                    // Move the Kinect in order to re-center the target
                                    SetKinectPosition(bearing);
                                    Thread.Sleep(1000);
                                }
                                m_LastVisualTrack = new TrackingResult() { Status = TrackStatus.Tracking, Bearing = bearing, Distance = Meters2Centimeters(skeleton.Position.Z, 0, m_UsingZoom), Quality = skeleton.Quality, TrackingID = skeleton.TrackingID };
                            }
                            else
                                m_LastVisualTrack = new TrackingResult() { Status = TrackStatus.NoJoint, Bearing = m_LastVisualTrack.Bearing, Distance = m_LastVisualTrack.Distance, Quality = m_LastVisualTrack.Quality, TrackingID = m_LastVisualTrack.TrackingID };
                        }
                        else
                            m_LastVisualTrack = new TrackingResult() { Status = TrackStatus.NoPerson, Bearing = m_LastVisualTrack.Bearing, Distance = m_LastVisualTrack.Distance, Quality = m_LastVisualTrack.Quality, TrackingID = m_LastVisualTrack.TrackingID };
                    }
                    else
                        m_LastVisualTrack = new TrackingResult() { Status = TrackStatus.NoFrame, Bearing = m_LastVisualTrack.Bearing, Distance = m_LastVisualTrack.Distance, Quality = m_LastVisualTrack.Quality, TrackingID = m_LastVisualTrack.TrackingID };

                    // Fire visual track event
                    if (VisualTrack != null)
                    {
                        m_LastVisualTrack.ResultType = TrackingMode.Visual;
                        VisualTrack(m_LastVisualTrack);
                    }
                }
            }
        }

        private void AudioTrackingWithSR()
        {
            // We need to run in high priority to avoid dropping samples 
            Thread.CurrentThread.Priority = ThreadPriority.Highest;

            // Instantiate the KinectAudioSource to do audio capture
            using (var source = new KinectAudioSource())
            {
                source.FeatureMode = true;
                source.AutomaticGainControl = false; // Important to turn this off for speech recognition
                source.SystemMode = SystemMode.OptibeamArrayOnly; // No AEC for this sample

                RecognizerInfo ri = SpeechRecognitionEngine.InstalledRecognizers().Where(r => r.Id == RecognizerId).FirstOrDefault();

                // Register for beam tracking change notifications
                source.BeamChanged += source_BeamChanged;

                using (var sre = new SpeechRecognitionEngine(ri.Id))
                {
                    var colors = new Choices();
                    colors.Add("red");
                    colors.Add("blue");
                    colors.Add("green");

                    var gb = new GrammarBuilder();
                    // Specify the culture to match the recognizer in case we are running in a different culture.                                 
                    gb.Culture = ri.Culture;
                    gb.Append(colors);

                    // Create the actual Grammar instance, and then load it into the speech recognizer.
                    var g = new Grammar(gb);
                    sre.LoadGrammar(g);
                    sre.SpeechRecognized += SreSpeechRecognized;
                    sre.SpeechHypothesized += SreSpeechHypothesized;
                    sre.SpeechRecognitionRejected += SreSpeechRecognitionRejected;

                    // Start capturing audio                               
                    using (var audioStream = source.Start())
                    {
                        sre.SetInputToAudioStream(audioStream,
                                                  new SpeechAudioFormatInfo(
                                                      EncodingFormat.Pcm, 16000, 16, 1,
                                                      32000, 2, null));

                        sre.RecognizeAsync(RecognizeMode.Multiple);

                        while (!m_StopAudioTracking)
                        {
                            // If we have high confidence, save the position
                            if (source.SoundSourcePositionConfidence > 0.9)
                            {
                                double deg = source.SoundSourcePosition * (180.0 / Math.PI);
                            }
                        }

                        sre.RecognizeAsyncCancel();
                        source.Stop();
                    }
                }
            }
        }

        private void AudioTracking()
        {
            // We need to run in high priority to avoid dropping samples 
            Thread.CurrentThread.Priority = ThreadPriority.Highest;

            // Instantiate the KinectAudioSource to do audio capture
            using (var source = new KinectAudioSource())
            {
                source.SystemMode = SystemMode.OptibeamArrayOnly;

                // Register for beam tracking change notifications
                source.BeamChanged += source_BeamChanged;

                // Start capturing audio                               
                using (var audioStream = source.Start())
                {
                    while (!m_StopAudioTracking)
                    {
                        // If we have high confidence, save the position
                        if (source.SoundSourcePositionConfidence > 0.9)
                        {
                            double deg = source.SoundSourcePosition * (180.0 / Math.PI);
                        }
                    }
                    source.Stop();
                }
            }
        }

        private void source_BeamChanged(object sender, BeamChangedEventArgs e)
        {
            if (m_TrackingMode == TrackingMode.Audio)
            {
                double degrees = e.Angle * (180.0 / Math.PI);
                int newPos = Convert.ToInt32(Math.Round(degrees, 0));

                if (m_ActiveTracking)
                {
                    // Compute bearing to target
                    int bearing = m_CurrentKinectPos - newPos;

                    SetKinectPosition(bearing);
                    Thread.Sleep(1000);
                }

                // Adjust to match visual tracking (0-180)
                if (newPos < 0)
                    newPos = Math.Abs(newPos) + 90;
                else
                    newPos = 90 - newPos;

                m_LastAudioTrack = new TrackingResult() { ResultType = TrackingMode.Audio, Bearing = newPos, Status = TrackStatus.Tracking };

                // Fire audio track event
                if (AudioTrack != null)
                    AudioTrack(m_LastAudioTrack);
            }
        }

        private void SreSpeechRecognitionRejected(object sender, SpeechRecognitionRejectedEventArgs e)
        {
            System.Diagnostics.Debug.WriteLine("Speech Rejected");
        }

        private void SreSpeechHypothesized(object sender, SpeechHypothesizedEventArgs e)
        {
            System.Diagnostics.Debug.WriteLine("Speech Hypothesized: " + e.Result.Text);
        }

        private void SreSpeechRecognized(object sender, SpeechRecognizedEventArgs e)
        {
            //This first release of the Kinect language pack doesn't have a reliable confidence model, so 
            //we don't use e.Result.Confidence here.
            System.Diagnostics.Debug.WriteLine("Speech Recognized: " + e.Result.Text);
        }

        private double Meters2Centimeters(float meters, int digits, bool usingZoom)
        {
            double cm = (double)meters * 100;
            if (usingZoom)
                return Math.Round(cm - (cm * .25), digits);
            else
                return Math.Round(cm, digits);
        }

        private double Meters2Inches(float meters, int digits, bool usingZoom)
        {
            double inches = (double)meters * 39.37;
            if (usingZoom)
                return Math.Round(inches - (inches * .25), digits);
            else
                return Math.Round(inches, digits);
        }

        #endregion
    }
}
