﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Research.Kinect.Nui;
using Coding4Fun.Kinect.Wpf;

namespace KSeeM1
{
    public class CorridorFinder
    {
        public struct ScanResult
        {
            public double X;
            public double Y;
            public double Distance { get; set; }
            public double ImageFrameWidth;
            public double ImageFrameHeight;
            public bool Hit;
        }

        private HotZoneSpec m_Corridor;

        public delegate void ScanHandler(ScanResult result);
        public event ScanHandler Scan;

        public CorridorFinder(double y, double width, double height, double minZ, double maxZ)
        {
            m_Corridor = new HotZoneSpec("Corridor", 0, y, width, height, minZ, maxZ);
        }

        public double? Process(ImageFrame imageFrame)
        {
            if (Scan != null)
                CallbackAllDepthData(imageFrame);

            const double step = 5.0;
            double offset = m_Corridor.Rectangle.Width / 2.0;
            double start = 40.0 - offset;
            double left = start - step;
            double right = start + step;
            double count = start / step;

            // Center
            HotZoneSpec centerCorridor = new HotZoneSpec("Center", start, m_Corridor.Rectangle.Y, m_Corridor.Rectangle.Width,
                m_Corridor.Rectangle.Height, m_Corridor.MinZ, m_Corridor.MaxZ);
            if (CheckCorridor(centerCorridor, imageFrame) == 0)
                return start + offset;

            // Scan left and right
            for (; count > 0; count--)
            {
                // Left
                HotZoneSpec leftCorridor = new HotZoneSpec("Left", left, m_Corridor.Rectangle.Y, m_Corridor.Rectangle.Width,
                    m_Corridor.Rectangle.Height, m_Corridor.MinZ, m_Corridor.MaxZ);
                if (CheckCorridor(leftCorridor, imageFrame) == 0)
                    return left + offset;

                // Right
                HotZoneSpec rightCorridor = new HotZoneSpec("Right", right, m_Corridor.Rectangle.Y, m_Corridor.Rectangle.Width,
                    m_Corridor.Rectangle.Height, m_Corridor.MinZ, m_Corridor.MaxZ);
                if (CheckCorridor(rightCorridor, imageFrame) == 0)
                    return right + offset;

                // Move to next step in the scan
                left -= step;
                right += step;
            }

            return null;
        }

        public HotZoneSpec Corridor
        {
            get { return m_Corridor; }
        }

        private double CheckCorridor(HotZoneSpec corridor, ImageFrame imageFrame)
        {
            int height = imageFrame.Image.Height;
            int width = imageFrame.Image.Width;

            // Depth data for each pixel
            Byte[] depthData = imageFrame.Image.Bits;

            // Scan all pixels and look for hits in the corridor
            var depthIndex = 0;
            for (var y = 0; y < height; y++)
            {
                var heightOffset = y * width;
                for (var x = 0; x < width; x++)
                {
                    var index = (x + heightOffset) * 4;
                    var distance = GetDistance(depthData[depthIndex], depthData[depthIndex + 1]);

                    bool hit = false;
                    if (corridor.Contains(x, y, distance))
                        hit = true;

                    if (Scan != null && hit)
                        Scan(new ScanResult() { X = x, Y = y, Distance = distance, Hit = hit, ImageFrameWidth = imageFrame.Image.Width, ImageFrameHeight = imageFrame.Image.Height });

                    // Move two bytes at a time
                    depthIndex += 2;
                }
            }

            // Return hit percentage
            return corridor.HitPercent;
        }

        private void CallbackAllDepthData(ImageFrame imageFrame)
        {
            int height = imageFrame.Image.Height;
            int width = imageFrame.Image.Width;

            // Depth data for each pixel
            Byte[] depthData = imageFrame.Image.Bits;

            // Scan all pixels and look for hits in the corridor
            var depthIndex = 0;
            for (var y = 0; y < height; y++)
            {
                var heightOffset = y * width;
                for (var x = 0; x < width; x++)
                {
                    var index = (x + heightOffset) * 4;
                    var distance = GetDistance(depthData[depthIndex], depthData[depthIndex + 1]);

                    Scan(new ScanResult() { X = x, Y = y, Distance = distance, Hit = false, ImageFrameWidth = imageFrame.Image.Width, ImageFrameHeight = imageFrame.Image.Height });

                    // Move two bytes at a time
                    depthIndex += 2;
                }
            }
        }

        private int GetDistance(byte firstFrame, byte secondFrame)
        {
            int distance = (int)firstFrame | secondFrame << 8;
            return distance;
        }
    }
}
