﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;
using AForge.Vision.GlyphRecognition;
using Microsoft.Research.Kinect.Nui;
using AForge;
using AForge.Math.Geometry;

namespace KSeeM1
{
    public class GlyphDetector
    {
        private GlyphRecognizer m_Recognizer;
        private double m_FOV;
        private double m_RobotOffset;

        public GlyphDetector(int glyphSize, int FieldOfView, int CameraScanWidth)
        {
            m_Recognizer = new GlyphRecognizer(glyphSize);
            m_Recognizer.GlyphDatabase = new GlyphDatabase(glyphSize);
            m_FOV = FieldOfView;
            m_RobotOffset = CameraScanWidth / 2;
        }

        public void AddGlyph(string glyphName, string glyphData)
        {
            Glyph glyph = new Glyph(glyphName, GlyphDataFromString(glyphData, m_Recognizer.GlyphSize));
            m_Recognizer.GlyphDatabase.Add(glyph);
        }

        public HotZoneSpec? LocateGlyph(string glyphName, List<HotZoneSpec> glyphHotzones)
        {
            foreach (HotZoneSpec hz in glyphHotzones)
            {
                // Return glyph's bearing if found
                if (hz.Name == glyphName)
                    return hz;
            }
            return null;
        }

        public HotZoneSpec? LocateGlyph(string glyphName, ImageFrame imageColor, ImageFrame imageDepth, double positionCamera)
        {
            List<HotZoneSpec> glyphHotzones = LocateAllGlyphs(imageColor, imageDepth, positionCamera);
            foreach (HotZoneSpec hz in glyphHotzones)
            {
                // Return glyph's bearing if found
                if (hz.Name == glyphName)
                    return hz;
            }
            return null;
        }

        public List<HotZoneSpec> LocateAllGlyphs(ImageFrame imageColor, ImageFrame imageDepth, double positionCamera)
        {
            // Prepare a bitmap
            Bitmap bitmap = PImageToBitmap(imageColor.Image);
            bitmap.RotateFlip(RotateFlipType.Rotate180FlipY);

            // Find glyphs
            List<ExtractedGlyphData> glyphs = m_Recognizer.FindGlyphs(bitmap);

            // Convert found and recognized glyphs into a list of hotzones
            List<HotZoneSpec> hotzones = new List<HotZoneSpec>();
            foreach (ExtractedGlyphData egd in glyphs)
            {
                IntPoint minXY, maxXY;
                PointsCloud.GetBoundingRectangle(egd.Quadrilateral, out minXY, out maxXY);
                if (egd.RecognizedGlyph != null)
                {
                    // This is a recognized glyph 
                    HotZoneSpec hz = new HotZoneSpec(egd.RecognizedGlyph.Name, minXY.X, minXY.Y, maxXY.X - minXY.X, maxXY.Y - minXY.Y, 0, 0);
                    hz.Confidence = Math.Round(egd.Confidence * 100.0, 0);

                    // Compute bearings relative to the camera and the robot
                    hz.CameraBearing = ComputeCameraBearing(hz, imageColor.Image.Width);
                    hz.RobotBearing = m_RobotOffset - (positionCamera + hz.CameraBearing);
                    
                    // Compute distance to glyph
                    if (imageDepth != null)
                        hz.Distance = GetHotzoneDepth(hz, imageColor, imageDepth);

                    hotzones.Add(hz);
                }
            }

            return hotzones;
        }

        #region Helpers

        private double GetHotzoneDepth(HotZoneSpec hz, ImageFrame imageColor, ImageFrame imageDepth)
        {
            int scale = imageColor.Image.Width / imageDepth.Image.Width;

            // Scale down since the depth data is 1/8 the resolution
            int depthX = (int)hz.Center.X / scale;
            int depthY = (int)hz.Center.Y / scale;

            // Compute the offeset in the depth data
            var heightOffset = depthY * imageDepth.Image.Width;
            var index = (depthX + heightOffset) * 2;

            // Return the distance
            return GetDistance(imageDepth.Image.Bits[index], imageDepth.Image.Bits[index + 1]);
        }

        private int GetDistance(byte firstFrame, byte secondFrame)
        {
            int distance = (int)firstFrame | secondFrame << 8;
            return distance;
        }

        private double ComputeCameraBearing(HotZoneSpec hz, int imageFrameWidth)
        {
            // Compute bearing relative to the camera and it's field of view and image width
            double X = hz.Rectangle.X + (hz.Rectangle.Width / 2);
            double width = imageFrameWidth / 2;
            return Math.Round((m_FOV / 2) * ((X - width) / width), 0);
        }

        private Bitmap PImageToBitmap(PlanarImage PImage)
        {
            Bitmap bmap = new Bitmap(PImage.Width, PImage.Height, PixelFormat.Format32bppRgb);
            BitmapData bmapdata = bmap.LockBits(new System.Drawing.Rectangle(0, 0, PImage.Width, PImage.Height),
                ImageLockMode.WriteOnly, bmap.PixelFormat);

            IntPtr ptr = bmapdata.Scan0;
            System.Runtime.InteropServices.Marshal.Copy(PImage.Bits, 0, ptr, PImage.Width * PImage.BytesPerPixel * PImage.Height);
            bmap.UnlockBits(bmapdata);

            return bmap;
        }

        private byte[,] GlyphDataFromString(string glyphStrData, int glyphSize)
        {
            byte[,] glyphData = new byte[glyphSize, glyphSize];

            for (int i = 0, k = 0; i < glyphSize; i++)
            {
                for (int j = 0; j < glyphSize; j++, k++)
                {
                    glyphData[i, j] = (byte)((glyphStrData[k] == '0') ? 0 : 1);
                }
            }

            return glyphData;
        }

        #endregion
    }
}
