﻿//
// K-See Mark 1
// Developed by Steve Norris (NorrisLabs).
// This is the main code for the K-See Kinect-based robot as descibed in Issue 33 of Robot Magazine.
// This application was developed in C# and uses the Windows Presentation Foundation (WPF) and Microsoft Kinect SDK.
// It was built to experiment with the Kinect and is still very much a work in progress.
// You may reuse all or any parts of this code to build your own Kinect-based robot application.
// This code is completely unsupported. NorrisLabs is not obligated or responsible for getting it to work with your robot.
//
// Current Status (as of 01/23/2012): 
// Glyph navigation and close-in obstacle avoidance is working.
// Eventually this code will be ported to my EmergeRT Robot Operation System.
//
// A more complete version may be available on my web site (http://www.norrislabs.com) in the future.
// 
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.IO.Ports;
using System.Windows.Threading;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Controls.Primitives;
using Microsoft.Research.Kinect.Nui;
using Coding4Fun.Kinect.Wpf;

namespace KSeeM1
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        private Runtime m_NUI;
        
        private RobotController m_KSeeController;
        private const int m_MaxDistanceIR = 80;
        private const int m_MaxDistancePing = 255;

        private DispatcherTimer m_timerBehaviors;

        private enum RobotMode
        {
            Inactive,
            NavigateRoute,
            FollowMe
        }
        private RobotMode m_Mode = RobotMode.Inactive;
        private int m_Speed = 1;

        private int m_PingNo = 0;

        // Glyph detection and tracking
        private Task m_TrackingTask;
        private CancellationTokenSource m_TrackingTaskCancel;

        private GlyphDetector m_GlyphDetector;
        private int m_LostGlyphCount;

        private string m_CurrentGlyphName = "U";
        private bool m_ActiveTracking = false;

        private ReaderWriterLockSlim m_GlyphLock = new ReaderWriterLockSlim();
        private ImageFrame m_ImageDepth;
        private ImageFrame m_ImageColor;
        private List<HotZoneSpec> m_GlyphHotzones;
        private HotZoneSpec? m_GlyphBearing = null;

        // Kinect depth specifications
        const float MinDepthDistance = 800;
        const float MaxDepthDistance = 4000;
        const float MaxDepthDistanceOffset = MaxDepthDistance - MinDepthDistance;

        // PIDs
        PID m_PidAvoid = new PID(2, 0, .2, 0, 100, 60, 100);
        PID m_PidGotoGlyph = new PID(1, 0, .1, -100, 100, 0, 100);

        // Route Navigation
        private struct NavInstruction
        {
            public string GlyphName { get; set; }
            public int Spin { get; set; }
        }
        private List<NavInstruction> m_Route;
        private int m_NICounter = 0;

        // Robot Behaviors
        private enum RobotBehavior
        {
            Inactive,
            SeeGlyph,
            Blocked,
            Move,
            Avoid,
            Navigator
        }
        private RobotBehavior m_LastBehavior;

        // Robot actions
        private enum RobotAction
        {
            NoAction,
            Stopped,
            MovForward,
            AvoidLeft,
            AvoidRight,
            Navigate
        }
        private RobotAction m_LastAction;

        // Frame rate
        private int m_TotalFrames = 0;
        private int m_LastFrames = 0;
        private DateTime m_LastTime = DateTime.MaxValue;

        private int m_TotalFrames2 = 0;
        private int m_LastFrames2 = 0;
        private DateTime m_LastTime2 = DateTime.MaxValue;
        private int m_TrackingFrameRate;

        public MainWindow()
        {
            InitializeComponent();

            try
            {
                Rect bounds = Properties.Settings.Default.WinPosition;
                this.Top = bounds.Top;
                this.Left = bounds.Left;
                this.Height = bounds.Height;
                this.Width = bounds.Width;
            }
            catch
            {
            }
        }

        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            txtNavRoute.Text = Properties.Settings.Default.Route;
            txtSpeedSelect.Text = Properties.Settings.Default.Speed.ToString();
            cbActiveTracking.IsChecked = true;

            // Load serial com ports and select the previous port
            try
            {
                List<string> ports = SerialPort.GetPortNames().ToList<string>();
                foreach (string port in ports)
                    cbComPort.Items.Add(port);

                if (ports.Count > 0)
                {
                    if (!string.IsNullOrWhiteSpace(Properties.Settings.Default.ComPort))
                        cbComPort.SelectedItem = Properties.Settings.Default.ComPort;
                    else
                        cbComPort.SelectedIndex = 0;
                }
                else
                {
                    MessageBox.Show("No serial ports available", "Com Port", MessageBoxButton.OK, MessageBoxImage.Exclamation);
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Com Port", MessageBoxButton.OK, MessageBoxImage.Exclamation);
            }

            // Setup Kinect runtime
            try
            {
                m_NUI = Runtime.Kinects[0];
                m_NUI.Initialize(RuntimeOptions.UseDepth | RuntimeOptions.UseColor);
            }
            catch(Exception ex)
            {
                MessageBox.Show("Runtime initialization failed.\n" + ex.Message, "Kinect", MessageBoxButton.OK, MessageBoxImage.Exclamation);
                return;
            }

            // Open Kinect data streams
            try
            {
                m_NUI.VideoStream.Open(ImageStreamType.Video, 2, ImageResolution.Resolution640x480, ImageType.Color);
                m_NUI.DepthStream.Open(ImageStreamType.Depth, 2, ImageResolution.Resolution80x60, ImageType.Depth);
            }
            catch (Exception ex)
            {
                MessageBox.Show("Failed to open data stream.\n" + ex.Message, "Kinect", MessageBoxButton.OK, MessageBoxImage.Exclamation);
                return;
            }

            // Set Kinect camera angle
            slTilt.Value = Properties.Settings.Default.CameraAngle;
            m_NUI.NuiCamera.ElevationAngle = (int)slTilt.Value;

            // Setup the detectors
            SetupGlyphDetector();

            // Setup to measure frame rates
            m_LastTime = m_LastTime2 = DateTime.Now;

            // Start the glyph tracking task
            m_TrackingTaskCancel = new CancellationTokenSource();
            m_TrackingTask = new Task(new Action(GlyphTracker), m_TrackingTaskCancel.Token);
            m_TrackingTask.Start();

            // Setup Process Behaviors timer
            m_timerBehaviors = new DispatcherTimer();
            m_timerBehaviors.Interval = TimeSpan.FromMilliseconds(40);   // Gives about 10 FPS
            m_timerBehaviors.Tick += new EventHandler(ProcessBehaviors);
            m_timerBehaviors.Start();
        }

        private void Window_Closing(object sender, System.ComponentModel.CancelEventArgs e)
        {
            Properties.Settings.Default.ComPort = (string)cbComPort.SelectedItem;
            Properties.Settings.Default.WinPosition = this.RestoreBounds;
            Properties.Settings.Default.Save();

            if(m_TrackingTaskCancel != null)
                m_TrackingTaskCancel.Cancel();

            if (m_KSeeController != null)
            {
                PowerDown();
                m_KSeeController.Dispose();
            }
        }

        private void btnClose_Click(object sender, RoutedEventArgs e)
        {
            this.Close();
        }

        private void cbComPort_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            if (m_KSeeController != null)
                m_KSeeController.Dispose();

            m_KSeeController = new RobotController((string)cbComPort.SelectedItem);
            if (m_KSeeController.UpdateSensorData(RobotController.Ping.Center))
            {
                m_KSeeController.MaxDistanceIR = m_MaxDistanceIR;
                m_KSeeController.MaxDistancePing = m_MaxDistancePing;
                m_KSeeController.SetKinectPosition(90);
            }
            else
            {
                MessageBox.Show("No serial communication with robot.", "Telemetry", MessageBoxButton.OK, MessageBoxImage.Error);
            }
        }

        private void cbUsingZoom_Click(object sender, RoutedEventArgs e)
        {

        }

        private void btnTilt_Click(object sender, RoutedEventArgs e)
        {
            m_NUI.NuiCamera.ElevationAngle = (int)slTilt.Value;
            Properties.Settings.Default.CameraAngle = slTilt.Value;
            Properties.Settings.Default.Save();
        }

        private void btnInactive_Click(object sender, RoutedEventArgs e)
        {
            txtStatus.Text = "Stopped";
            btnInactive.Focus();
            btnNavRoute.IsChecked = false;
            btnFollowMe.IsChecked = false;
            m_Mode = RobotMode.Inactive;
        }

        private void btnNavRoute_Click(object sender, RoutedEventArgs e)
        {
            if (SaveMotionParameters())
            {
                btnNavRoute.Focus();
                btnFollowMe.IsChecked = false;
                m_Mode = RobotMode.NavigateRoute;
                ExecuteNavInstruction(true);
            }
        }

        private void btnFollowMe_Click(object sender, RoutedEventArgs e)
        {
            if (SaveMotionParameters())
            {
                btnNavRoute.IsChecked = false;
                btnFollowMe.Focus();
                m_Mode = RobotMode.FollowMe;
            }
        }

        private void cbActiveTracking_Checked(object sender, RoutedEventArgs e)
        {
            SetActiveTracking(true);
        }

        private void cbActiveTracking_Unchecked(object sender, RoutedEventArgs e)
        {
            if(m_KSeeController != null)
                m_KSeeController.SetKinectPosition(90);
            SetActiveTracking(false);
        }

        private void Window_PreviewKeyDown(object sender, KeyEventArgs e)
        {
            if (e.Key == Key.Space || e.Key == Key.F1)
            {
                btnInactive_Click(null, null);
            }

            if (e.Key == Key.F2)
            {
                btnNavRoute_Click(null, null);
                btnNavRoute.IsChecked = true;
            }

            if (e.Key == Key.F3)
            {
                btnFollowMe_Click(null, null);
                btnFollowMe.IsChecked = true;
            }

            if (e.Key == Key.F12)
            {
                this.Close();
            }
        }

        private void ProcessBehaviors(object sender, EventArgs e)
        {
            // Get sensor data
            GetTelemetry((RobotController.Ping)m_PingNo);
            m_PingNo += 1;
            if (m_PingNo > 2)
                m_PingNo = 1;

            // Display sensor data based on selected tab
            if (((TabItem)tabControl.SelectedItem).Name == "tiTelemetryData")
            {
                DisplayTelemetryData();
            }
            else if (((TabItem)tabControl.SelectedItem).Name == "tiTelemetryGraphic")
            {
                DisplayTelemetryGraphic();
            }
            else if (((TabItem)tabControl.SelectedItem).Name == "tiTracking")
            {
                // Display Kinect data streams
                m_GlyphLock.EnterReadLock();

                // Depth data stream
                if (m_ImageDepth != null)
                    DisplayDepthFrame(m_ImageDepth);

                // RGB data stream
                if (m_ImageColor != null)
                   DisplayColorFrame(m_ImageColor, m_ImageDepth);

                m_GlyphLock.ExitReadLock();
            }

            txtStatus.Text = m_LastBehavior.ToString() + ", " + m_LastAction.ToString();
            
            // Execute Behaviors
            switch (m_Mode)
            {
                case RobotMode.NavigateRoute:
                    NavigateStack();
                    break;

                case RobotMode.FollowMe:
                    break;

                default:
                    PowerDown();
                    break;
            }

            // Calculate and display frame rates per second
            DisplayFrameRates();
        }

        #region Behavior Stacks (Modes)

        private void NavigateStack()
        {
            // Find the current active glyph
            if (SeeGlyph())
                return;

            // Are we blocked?
            if (Blocked(true))
                return;

            // If not moving then get going
            if (Move())
                return;

            // Avoid left/right obstacles
            if (Avoid())
                return;

            // Navigate route
            NavRoute();
        }

        #endregion

        #region Behaviors

        private bool SeeGlyph()
        {
            m_LastBehavior = RobotBehavior.SeeGlyph;
            m_LastAction = RobotAction.NoAction;

            if (!GetGlyphBearing().HasValue)
            {
                m_LostGlyphCount++;
                if (m_LostGlyphCount > 3)
                {
                    // Are we in motion?
                    m_KSeeController.UpdateStatus();
                    if (m_KSeeController.CurrentDirection != RobotController.Direction.Halt)
                    {
                        // Halt
                        m_KSeeController.Halt();
                        Thread.Sleep(1000);
                        m_KSeeController.SoftReset();

                        // Find glyph
                        if (FindGlyph(m_CurrentGlyphName))
                        {
                            m_LostGlyphCount = 0;
                            return false;
                        }
                    }

                    // Inhibit
                    m_LastAction = RobotAction.Stopped;
                    return true;
                }

                // Don't inhibit just yet
                return false;
            }

            m_LostGlyphCount = 0;
            return false;
        }

        private bool Blocked(bool backup)
        {
            m_LastBehavior = RobotBehavior.Blocked;
            m_LastAction = RobotAction.NoAction;

            if ((m_KSeeController.PingLeft < 25 && m_KSeeController.PingRight < 25) || m_KSeeController.IRCenter < 20)
            {
                // Halt and inhibit
                m_KSeeController.EmergencyHalt();
                if (backup)
                    m_KSeeController.MoveReverse(12, true);

                // Spin in the direction that the Kinect was looking
                if (m_KSeeController.KinectPosition > 90)
                    m_KSeeController.SpinCW(30, true);
                else
                    m_KSeeController.SpinCCW(30, true);
                m_KSeeController.SetKinectPosition(90);

                Thread.Sleep(1000);

                m_LastAction = RobotAction.Stopped;
                return true;
            }

            return false;
        }

        private bool Move()
        {
            m_LastBehavior = RobotBehavior.Move;
            m_LastAction = RobotAction.MovForward;

            // Power up
            if (!m_KSeeController.DrivePower)
            {
                m_KSeeController.SetPower(true);
                System.Threading.Thread.Sleep(2500);
                m_KSeeController.UpdateStatus();
            }

            // Move forward
            m_KSeeController.UpdateStatus();
            if (m_KSeeController.CurrentDirection == RobotController.Direction.Halt)
                m_KSeeController.MoveForward(m_Speed, 2000, false);

            return false;
        }

        private bool Avoid()
        {
            m_LastBehavior = RobotBehavior.Avoid;
            m_LastAction = RobotAction.NoAction;

            // Suspend sensors as we approach the glyph
            HotZoneSpec? glyphBearing = GetGlyphBearing();
            if (glyphBearing.HasValue)
            {
                if (glyphBearing.Value.Distance < 1000)
                    return false;
            }

            int ldist = Math.Min(m_KSeeController.PingLeft, m_KSeeController.IRLeft);
            int rdist = Math.Min(m_KSeeController.PingRight, m_KSeeController.IRRight);
            
            double? bias;
            int turn;
            if (ldist <= 60 && ldist < rdist)
            {
                // Right turn
                bias = m_PidAvoid.Compute(ldist);
                if (bias.HasValue && bias >= 1)
                {
                    turn = Convert.ToInt32(Math.Round((double)bias, 0));
                    m_KSeeController.TurnRight(turn);
                    m_LastAction = RobotAction.AvoidRight;
                    return true;
                }
            }
            else if (rdist <= 60)
            {
                // Left turn
                bias = m_PidAvoid.Compute(rdist);
                if (bias.HasValue && bias >= 1)
                {
                    turn = Convert.ToInt32(Math.Round((double)bias, 0));
                    m_KSeeController.TurnLeft(turn);
                    m_LastAction = RobotAction.AvoidLeft;
                    return true;
                }
            }

            return false;
        }

        private bool NavRoute()
        {
            m_LastBehavior = RobotBehavior.Navigator;
            m_LastAction = RobotAction.Navigate;

            m_GlyphLock.EnterReadLock();
            HotZoneSpec? glyphBearing = m_GlyphBearing;
            m_GlyphLock.ExitReadLock();

            if (glyphBearing.HasValue)
            {
                m_LastAction = RobotAction.Navigate;
                if (glyphBearing.Value.Distance < 900)
                {
                    m_KSeeController.TurnRight(1);
                    m_KSeeController.Halt();
                    Thread.Sleep(1000);

                    ExecuteNavInstruction(false);
                }
                else
                    GotoGlyph(glyphBearing.Value);
            }

            return false;
        }

        #endregion

        #region Helpers

        private HotZoneSpec? GetGlyphBearing()
        {
            m_GlyphLock.EnterReadLock();
            HotZoneSpec? glyphBearing = m_GlyphBearing;
            m_GlyphLock.ExitReadLock();
            return glyphBearing;
        }

        private bool FindGlyph(string nameGlyph)
        {
            // Turn off active tracking if it is on
            bool tracking = m_ActiveTracking;
            SetActiveTracking(false);

            // Look again where we were looking
            Thread.Sleep(1000);
            if (m_GlyphBearing.HasValue)
            {
                SetActiveTracking(tracking);
                return true;
            }

            // Now conduct a scanning search
            int[] searchPattern = { 30, 60, 90, 120, 150 };
            foreach (int pos in searchPattern)
            {
                m_KSeeController.SetKinectPosition(pos);
                Thread.Sleep(1000);
                if (m_GlyphBearing.HasValue)
                {
                    SetActiveTracking(tracking);
                    return true;
                }
            }

            // Damn! We did not find the glyph
            m_KSeeController.SetKinectPosition(90);
            SetActiveTracking(tracking);
            return false;
        }

        private void ExecuteNavInstruction(bool startup)
        {
            if (startup)
                m_NICounter = 0;

            // Get nav instruction
            NavInstruction ni = m_Route[m_NICounter];

            // Set next glyph name
            SetActiveGlyph(ni.GlyphName);

            // Spin the robot
            if (!startup)
            {
                if (ni.Spin < 0)
                    m_KSeeController.SpinCCW(Math.Abs(ni.Spin), true);
                else
                    m_KSeeController.SpinCW(ni.Spin, true);
            }

            // Face Kinect forward
            m_KSeeController.SetKinectPosition(90);

            Thread.Sleep(1000);

            // Point to next instruction
            m_NICounter += 1;
            if (m_NICounter == m_Route.Count)
                m_NICounter = 0;
        }

        private void SetActiveGlyph(string nameGlyph)
        {
            m_GlyphLock.EnterWriteLock();
            m_CurrentGlyphName = nameGlyph;
            m_GlyphLock.ExitWriteLock();

            txtActiveGlyph.Text = nameGlyph;
        }

        private void SetupGlyphDetector()
        {
            m_GlyphDetector = new GlyphDetector(5, 60, 180);
            m_GlyphDetector.AddGlyph("U", "0000001110010000111000000");
            m_GlyphDetector.AddGlyph("T", "0000001110001000010000000");
        }

        private bool GotoGlyph(HotZoneSpec glyphBearing)
        {
            double? bias = m_PidGotoGlyph.Compute(glyphBearing.RobotBearing);
            if (bias.HasValue)
            {
                int turn = Convert.ToInt32(Math.Round(Math.Abs((double)bias), 0));
                if (bias < 0)
                    m_KSeeController.TurnLeft(turn);
                else if (bias > 0)
                    m_KSeeController.TurnRight(turn);
                else
                    m_KSeeController.TurnRight(1);
                return true;
            }
            return false;
        }

        private void GlyphTracker()
        {
            while (true)
            {
                // Check for task cancellation
                if (m_TrackingTaskCancel.Token.IsCancellationRequested)
                    break;

                if (m_KSeeController != null)
                {
                    m_GlyphLock.EnterWriteLock();

                    // Get Kinect data frames
                    m_ImageDepth = m_NUI.DepthStream.GetNextFrame(100);
                    m_ImageColor = m_NUI.VideoStream.GetNextFrame(100);

                    // Look for the active glyph in data streams
                    if (m_ImageColor != null)
                    {
                        m_GlyphHotzones = m_GlyphDetector.LocateAllGlyphs(m_ImageColor, m_ImageDepth, m_KSeeController.KinectPosition);
                        m_GlyphBearing = m_GlyphDetector.LocateGlyph(m_CurrentGlyphName, m_GlyphHotzones);
                        if (m_ActiveTracking && m_GlyphBearing.HasValue)
                        {
                            // Move the Kinect to re-center the glyph
                            int diff = (int)Math.Round((double)m_GlyphBearing.Value.CameraBearing, 0);
                            int adiff = Math.Abs(diff);
                            if (adiff > 10)
                            {
                                // Reposition the servo
                                int newPos = m_KSeeController.KinectPosition + diff;
                                m_KSeeController.SetKinectPosition(newPos);
                                Thread.Sleep(200);

                                // Throw away movement blurred frames
                                for (int i = 0; i < 2; i++)
                                {
                                    m_NUI.DepthStream.GetNextFrame(100);
                                    m_NUI.VideoStream.GetNextFrame(100);
                                }
                            }
                        }
                    }

                    m_GlyphLock.ExitWriteLock();
                    Thread.Sleep(20);
                }

                CalcTrackingFrameRate();
            }
        }

        private void SetActiveTracking(bool isOn)
        {
            m_GlyphLock.EnterWriteLock();
            m_ActiveTracking = isOn;
            m_GlyphLock.ExitWriteLock();
        }

        private bool SaveMotionParameters()
        {
            int speedSelect;
            if (int.TryParse(txtSpeedSelect.Text, out speedSelect))
            {
                m_Speed = speedSelect;
                Properties.Settings.Default.Speed = speedSelect;
            }
            else
            {
                MessageBox.Show("Invalid Speed.", "Invalid Motion Parameters", MessageBoxButton.OK, MessageBoxImage.Asterisk);
                return false;
            }

            // Parse route
            m_Route = new List<NavInstruction>();
            try
            {
                string[] items = txtNavRoute.Text.Split(',');
                for (int i = 0; i < items.Length; i += 2)
                {
                    NavInstruction ni = new NavInstruction { GlyphName = items[i].Trim(), Spin = int.Parse(items[i + 1].Trim()) };
                    m_Route.Add(ni);
                }
                Properties.Settings.Default.Route = txtNavRoute.Text;
            }
            catch (Exception)
            {
                MessageBox.Show("Invalid Route.", "Invalid Motion Parameters", MessageBoxButton.OK, MessageBoxImage.Asterisk);
                return false;
            }

            return true;
        }

        private void PowerDown()
        {
            m_KSeeController.EmergencyHalt();
            m_KSeeController.SetPower(false);
        }

        private void GetTelemetry(RobotController.Ping ping)
        {
            if (m_KSeeController != null)
            {
                m_KSeeController.UpdateSensorData(ping);
                m_KSeeController.UpdateStatus();
            }
        }

        private void DisplayTelemetryData()
        {
            if (m_GlyphBearing == null)
                txtGlyphBearing.Text = "No Glyph";
            else
                txtGlyphBearing.Text = m_GlyphBearing.Value.RobotBearing.ToString() + ", " + m_GlyphBearing.Value.Distance.ToString();

            if (m_KSeeController != null)
            {
                txtLeft.Text = m_KSeeController.IRLeft.ToString() + "/" + m_KSeeController.PingLeft.ToString();
                txtCenter.Text = m_KSeeController.IRCenter.ToString();
                txtRight.Text = m_KSeeController.IRRight.ToString() + "/" + m_KSeeController.PingRight.ToString();

                txtDrivePower.Text = m_KSeeController.DrivePower ? "On" : "Off";
                txtSpeed.Text = m_KSeeController.CurrentSpeed.ToString();
                txtDirection.Text = m_KSeeController.CurrentDirection.ToString();
                txtWheelPosLeft.Text = m_KSeeController.WheelPositionLeft.ToString();
                txtWheelPosRight.Text = m_KSeeController.WheelPositionRight.ToString();
                txtHasArrived.Text = m_KSeeController.HasArrived ? "Yes" : "No";
            }
        }

        private void DisplayTelemetryGraphic()
        {
            if (m_KSeeController != null)
            {
                if (m_GlyphBearing == null)
                    SetThumbData(slBearing, Brushes.White, Brushes.Red, m_GlyphBearing);
                else
                    SetThumbData(slBearing, Brushes.Black, Brushes.Lime, m_GlyphBearing);

                pbLeftIR.Maximum = m_KSeeController.MaxDistanceIR;
                pbLeftPing.Maximum = m_KSeeController.MaxDistanceIR;
                pbCenterIR.Maximum = m_KSeeController.MaxDistanceIR;
                pbRightPing.Maximum = m_KSeeController.MaxDistanceIR;
                pbRightIR.Maximum = m_KSeeController.MaxDistanceIR;

                txtLeftIR.Text = m_KSeeController.IRLeft.ToString();
                txtLeftPing.Text = m_KSeeController.PingLeft.ToString();
                txtCenterIR.Text = m_KSeeController.IRCenter.ToString();
                txtRightPing.Text = m_KSeeController.PingRight.ToString();
                txtRightIR.Text = m_KSeeController.IRRight.ToString();

                pbLeftIR.Value = m_KSeeController.IRLeft;
                pbLeftIR.Foreground = m_KSeeController.IRLeft < 30 ? Brushes.Yellow : Brushes.Lime;
 
                pbLeftPing.Value = m_KSeeController.PingLeft;
                pbLeftPing.Foreground = m_KSeeController.PingLeft < 30 ? Brushes.Yellow : Brushes.Lime;
                
                pbCenterIR.Value = m_KSeeController.IRCenter;
                pbCenterIR.Foreground = m_KSeeController.IRCenter < 30 ? Brushes.Yellow : Brushes.Lime;
                
                pbRightPing.Value = m_KSeeController.PingRight;
                pbRightPing.Foreground = m_KSeeController.PingRight < 30 ? Brushes.Yellow : Brushes.Lime;

                pbRightIR.Value = m_KSeeController.IRRight;
                pbRightIR.Foreground = m_KSeeController.IRRight < 30 ? Brushes.Yellow : Brushes.Lime;
            }
        }

        private void DisplayDepthFrame(ImageFrame imageDepth)
        {
            // Build a colored image from depth data
            Byte[] colorFrame = BuildColorDepthFrame(imageDepth);

            // Display colored depth image
            PlanarImage image = imageDepth.Image;
            imgDepth.Source = BitmapSource.Create(image.Width, image.Height, 96, 96, PixelFormats.Bgr32, null,
                colorFrame, image.Width * PixelFormats.Bgr32.BitsPerPixel / 8);
        }

        private void DisplayColorFrame(ImageFrame imageColor, ImageFrame imageDepth)
        {
            // Display RGB image
            imgVideo.Source = imageColor.ToBitmapSource();
            ClearCanvas(cvsVideo);

            // Display any found glyphs
            foreach (HotZoneSpec hz in m_GlyphHotzones)
            {
                DrawHotZone(cvsVideo, .5, Brushes.Yellow, hz);
                DrawBearing(cvsVideo, .5, Brushes.Lime, hz);
            }
        }

        private Byte[] BuildColorDepthFrame(ImageFrame imageFrame)
        {
            int height = imageFrame.Image.Height;
            int width = imageFrame.Image.Width;

            // Build new ColorFrame that will contain colored depth information for all pixels in image
            // This frame will be filled in by the callback from the CorridorFinder
            // Height x Width x 4 (Red, Green, Blue, empty byte)
            Byte[] colorFrame = new byte[height * width * 4];

            // Hardcoded locations to Blue, Green, Red (BGR) index positions       
            const int BlueIndex = 0;
            const int GreenIndex = 1;
            const int RedIndex = 2;

            var depthIndex = 0;
            for (var y = 0; y < height; y++)
            {
                var heightOffset = y * width;
                for (var x = 0; x < width; x++)
                {
                    var index = (x + heightOffset) * 4;
                    var distance = GetDistance(imageFrame.Image.Bits[depthIndex], imageFrame.Image.Bits[depthIndex + 1]);

                    // Set pixel color based on distance or in a hot zone
                    if (distance == 0)
                    {
                        // Unknown distance, show in dark blue
                        colorFrame[index + BlueIndex] = 64;
                        colorFrame[index + GreenIndex] = 0;
                        colorFrame[index + RedIndex] = 0;
                    }
                    else
                    {
                        // Use gray scale coloring for monochromatic histogram
                        var intensity = CalcIntensityFromDepth(distance);
                        colorFrame[index + BlueIndex] = intensity;
                        colorFrame[index + GreenIndex] = intensity;
                        colorFrame[index + RedIndex] = intensity;
                    }

                    // Move two bytes at a time
                    depthIndex += 2;
                }
            }

            return colorFrame;
        }

        public void DrawHotZone(Canvas canvas, double scale, Brush stroke, HotZoneSpec hz)
        {
            System.Windows.Shapes.Rectangle rect = new System.Windows.Shapes.Rectangle();
            rect.Stroke = stroke;

            rect.Width = hz.Rectangle.Width * scale;
            rect.Height = hz.Rectangle.Height * scale;
            Canvas.SetLeft(rect, hz.Rectangle.Left * scale);
            Canvas.SetTop(rect, hz.Rectangle.Top * scale);
            canvas.Children.Add(rect);

            TextBlock tb = new TextBlock();
            tb.Text = hz.Name;
            tb.Foreground = stroke;
            tb.Width = rect.Width;
            tb.Height = 44;
            tb.TextWrapping = TextWrapping.Wrap;
            tb.TextAlignment = TextAlignment.Center;
            Canvas.SetLeft(tb, hz.Rectangle.Left * scale);
            Canvas.SetTop(tb, (hz.Rectangle.Top * scale) + ((rect.Height - tb.Height) / 2));
            canvas.Children.Add(tb);
        }

        public void DrawBearing(Canvas canvas, double scale, Brush stroke, HotZoneSpec hz)
        {
            System.Windows.Shapes.Line line = new System.Windows.Shapes.Line();
            line.Stroke = stroke;

            double X = hz.Rectangle.X + (hz.Rectangle.Width / 2);

            line.X1 = 1;
            line.X2 = 1;
            line.Y1 = 1;
            line.Y2 = 240;
            line.HorizontalAlignment = HorizontalAlignment.Left;
            line.VerticalAlignment = VerticalAlignment.Center;
            line.StrokeThickness = 1;
            Canvas.SetLeft(line, X * scale);
            Canvas.SetTop(line, 0);
            canvas.Children.Add(line);

            TextBlock tb = new TextBlock();
            tb.Text = hz.RobotBearing + ", " + hz.Distance.ToString();
            tb.Foreground = stroke;
            tb.Width = 60;
            tb.Height = 44;
            tb.TextWrapping = TextWrapping.Wrap;
            tb.TextAlignment = TextAlignment.Center;
            Canvas.SetLeft(tb, X * scale);
            Canvas.SetTop(tb, 220);
            canvas.Children.Add(tb);

        }

        private void ClearCanvas(Canvas canvas)
        {
            canvas.Children.Clear();
        }

        private void DisplayFrameRates()
        {
            // Calculate and display frames per second
            ++m_TotalFrames;
            DateTime cur = DateTime.Now;
            if (cur.Subtract(m_LastTime) > TimeSpan.FromSeconds(1))
            {
                int frameDiff = m_TotalFrames - m_LastFrames;
                m_LastFrames = m_TotalFrames;
                m_LastTime = cur;
                txtFrameRate.Text = frameDiff.ToString() + "/" + m_TrackingFrameRate.ToString();
            }
        }

        private void CalcTrackingFrameRate()
        {
            // Calculate and display frames per second
            ++m_TotalFrames2;
            DateTime cur = DateTime.Now;
            if (cur.Subtract(m_LastTime2) > TimeSpan.FromSeconds(1))
            {
                int frameDiff = m_TotalFrames2 - m_LastFrames2;
                m_LastFrames2 = m_TotalFrames2;
                m_LastTime2 = cur;
                m_TrackingFrameRate = frameDiff;
            }
        }

        private void SetThumbData(Slider slider, Brush foreground, Brush background, HotZoneSpec? bearing)
        {
            if (bearing.HasValue)
                slider.Value = bearing.Value.RobotBearing;
            else
                slider.Value = 0;

            var track = slider.Template.FindName("PART_Track", slider) as Track;
            var thumb = track.Thumb;

            var el = thumb.Template.FindName("Ellipse", thumb) as Ellipse;
            el.Fill = background;

            var txt = thumb.Template.FindName("Text", thumb) as TextBlock;
            if (bearing.HasValue)
            {
                txt.Text = bearing.Value.RobotBearing.ToString();
                txtBearing.Text = bearing.Value.RobotBearing.ToString() + ", " + bearing.Value.Distance + ", " + bearing.Value.Name;
            }
            else
            {
                txt.Text = "?";
                txtBearing.Text = "No Glyph";
            }

            txt.Foreground = foreground;
            txt.Background = background;
        }

        private int GetDistance(byte firstFrame, byte secondFrame)
        {
            int distance = (int)firstFrame | secondFrame << 8;
            return distance;
        }

        private byte CalcIntensityFromDepth(double distance)
        {
            // formula for calculating monochrome intensity for histogram
            if (distance == 0)
                return 0;
            else
                return (byte)(255 - (255 * Math.Max(distance - MinDepthDistance, 0) / (MaxDepthDistanceOffset)));
        }

        private string DoubleToString(double? number)
        {
            if (number.HasValue)
                return number.ToString();
            else
                return "NA";
        }

        #endregion
    }
}
