﻿/**********************************************************************************************
 *
 * PID Library - Ported by Steve Norris to C# from the Arduino Version 1.0 
 *  written by Brett Beauregard
 *
 **********************************************************************************************/
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace KSeeM1
{
    public class PID
    {
        public enum Direction
        {
            Direct,
            Reverse
        }

        private const long TicksPerMillisecond = 10000;

	    double m_Kp;                  // (P)roportional Tuning Parameter
        double m_Ki;                  // (I)ntegral Tuning Parameter
        double m_Kd;                  // (D)erivative Tuning Parameter

        double m_SetPoint;

	    Direction m_ControllerDirection;

	    long m_LastTime;
	    double m_ITerm, m_LastInput;

	    long m_SampleTime;
	    double m_OutMin, m_OutMax;

        /* Constructor *************************************************************
         **************************************************************************/
        public PID(double Kp, double Ki, double Kd, double MinLimit, double MaxLimit, double SetPoint, long sampleTime)
        {
            m_SetPoint = SetPoint;
            m_SampleTime = sampleTime * TicksPerMillisecond;
            SetTunings(Kp, Ki, Kd);
            SetOutputLimits(MinLimit, MaxLimit);

            SetControllerDirection(Direction.Direct);
            m_LastTime = DateTime.Now.Ticks - m_SampleTime;				
        }

        /* Compute **********************************************************************
         *   The function will decide whether a new PID Output needs to be computed based
         *   on the sample time
         *******************************************************************************/
        public double? Compute(double myInput)
        {
           long now = DateTime.Now.Ticks;
           long timeChange = (now - m_LastTime);
           if(timeChange >= m_SampleTime)
           {
              // Compute all the working error variables
	          double input = myInput;
              double error = m_SetPoint - input;
              m_ITerm += (m_Ki * error);

              if(m_ITerm > m_OutMax)
                  m_ITerm = m_OutMax;
              else if(m_ITerm < m_OutMin)
                  m_ITerm = m_OutMin;
              
               double dInput = (input - m_LastInput);
 
              // Compute PID Output
              double output = (m_Kp * error) + m_ITerm - (m_Kd * dInput);
      
	          if(output > m_OutMax)
                  output = m_OutMax;
              else if(output < m_OutMin)
                  output = m_OutMin;
	  
              // Save for the next time
              m_LastInput = input;
              m_LastTime = now;

              return output;
           }

           return null;
        }

        /* SetOutputLimits *********************************************************
         **************************************************************************/
        public void SetOutputLimits(double Min, double Max)
        {
            if (Min >= Max)
                return;

            m_OutMin = Min;
            m_OutMax = Max;

            if (m_ITerm > m_OutMax)
                m_ITerm = m_OutMax;
            else if (m_ITerm < m_OutMin)
                m_ITerm = m_OutMin;
        }

        /* SetTunings *****************************************************************
         *  This function allows the controller's dynamic performance to be adjusted. 
         *  It's called from the constructor, but tunings can also be adjusted on the fly.
         ******************************************************************************/
        public void SetTunings(double Kp, double Ki, double Kd)
        {
            if (Kp < 0 || Ki < 0 || Kd < 0)
                return;

            double SampleTimeInSec = ((double)m_SampleTime) / (1000 * TicksPerMillisecond);
            m_Kp = Kp;
            m_Ki = Ki * SampleTimeInSec;
            m_Kd = Kd / SampleTimeInSec;

            if (m_ControllerDirection == Direction.Reverse)
            {
                m_Kp = (0 - m_Kp);
                m_Ki = (0 - m_Ki);
                m_Kd = (0 - m_Kd);
            }
        }

        /* SetSampleTime *************************************************************
        *   Sets the period, in milliseconds, at which the calculation is performed.
        *   This is also called from the constructor.
        ******************************************************************************/
        public void SetSampleTime(long NewSampleTime)
        {
           if (NewSampleTime > 0)
           {
              double ratio  = (double)(NewSampleTime * TicksPerMillisecond) / (double)m_SampleTime;
              m_Ki *= ratio;
              m_Kd /= ratio;
              m_SampleTime = NewSampleTime;
           }
        }

        /* SetControllerDirection *****************************************************
         *  The PID will either be connected to a DIRECT acting process (+Output leads 
         *  to +Input) or a REVERSE acting process (+Output leads to -Input.)
         *  This is also called from the constructor.
         ******************************************************************************/
        public void SetControllerDirection(Direction dir)
        {
           if(dir != m_ControllerDirection)
           {
	          m_Kp = (0 - m_Kp);
              m_Ki = (0 - m_Ki);
              m_Kd = (0 - m_Kd);
           }

           m_ControllerDirection = dir;
        }
    }
}
