﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO.Ports;
using System.Threading;

namespace KSeeM1
{
    public class RobotController : IDisposable
    {
        private SerialPort m_Port;
        private bool m_Disposed = false;

        public enum Direction
        {
            Halt,
            Forward,
            Reverse,
            SpinCW,
            SpinCCW
        }

        public enum Ping
        {
            Center = 0,
            Left = 1,
            Right = 2,
            None = 3
        }

        private int m_MaxDistanceIR = 80;
        private int m_MaxDistancePing = 255;

        // Status
        private int m_KinectPos;
        private bool m_DrivePower;
        private int m_Speed;
        private Direction m_Direction;
        private int m_WheelPosLeft;
        private int m_WheelPosRight;
        private bool m_HasArrived;

        // Sensors
        private int m_IRLeft;
        private int m_IRCenter;
        private int m_IRRight;
        private int m_PingLeft;
        private int m_PingCenter;
        private int m_PingRight;

        public RobotController(string portName)
        {
            SetComPort(portName);
        }

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        private void Dispose(bool disposing)
        {
            // Check to see if Dispose has already been called.
            if (!this.m_Disposed)
            {
                // If disposing equals true, dispose all managed 
                // and unmanaged resources.
                if (disposing)
                {
                    // Dispose managed resources.
                    m_Port.Close();
                }

                // Call the appropriate methods to clean up 
                // unmanaged resources here.
                // If disposing is false, 
                // only the following code is executed.
            }
            m_Disposed = true;
        }

        public void SetPower(bool IsOn)
        {
            if(IsOn)
                SendCommand("P1");
            else
                SendCommand("p1");
        }

        public void SetSpeed(int speed)
        {
            SendCommand("S" + speed.ToString());
        }

        public void MoveForward()
        {
            SendCommand("FW");
        }

        public void MoveForward(int distance, bool wait)
        {
            SendCommand("FX" + distance.ToString());
            if (wait)
                WaitArrived();
        }

        public void MoveForward(int speed, int distance, bool wait)
        {
            SetSpeed(speed);
            MoveForward(distance, wait);
        }

        public void MoveReverse()
        {
            SendCommand("BK");
        }

        public void MoveReverse(int distance, bool wait)
        {
            SendCommand("BX" + distance.ToString());
            if (wait)
                WaitArrived();
        }

        public void MoveReverse(int speed, int distance, bool wait)
        {
            SetSpeed(speed);
            MoveReverse(distance, wait);
        }

        public void Halt()
        {
            SendCommand("HL");
        }

        public void EmergencyHalt()
        {
            SendCommand("EH");
        }

        public void SoftReset()
        {
            EmergencyHalt();
        }

        public void TurnLeft(int bias)
        {
            SendCommand("LX" + bias.ToString());
        }

        public void TurnRight(int bias)
        {
            SendCommand("RX" + bias.ToString());
        }

        public void SpinCW()
        {
            SendCommand("RT");
        }

        public void SpinCW(int degrees, bool wait)
        {
            SendCommand("RX" + degrees.ToString());
            if(wait)
                WaitArrived();
        }

        public void SpinCCW()
        {
            SendCommand("LF");
        }

        public void SpinCCW(int degrees, bool wait)
        {
            SendCommand("LX" + degrees.ToString());
            if (wait)
                WaitArrived();
        }

        public void SetKinectPosition(int pos)
        {
            if(pos >= 0 && pos <= 180)
                SendCommand("KX" + pos.ToString());
            m_KinectPos = pos;
        }

        public bool UpdateSensorData(Ping ping)
        {
            string frame;
            switch (ping)
            {
                case Ping.Center:
                    frame = "C";
                    break;

                case Ping.Left:
                    frame = "L";
                    break;

                case Ping.Right:
                    frame = "R";
                    break;

                default:
                    frame = "N";    // No Ping
                    break;
            }

            List<string> dataItems = GetTelemetry(frame);
            if (dataItems != null && dataItems.Count == 4)
            {
                m_IRLeft = int.Parse(dataItems[0]);
                m_IRCenter = int.Parse(dataItems[1]);
                m_IRRight = int.Parse(dataItems[2]);

                switch (ping)
                {
                    case Ping.Center:
                        m_PingCenter = int.Parse(dataItems[3]);
                        break;

                    case Ping.Left:
                        m_PingLeft = int.Parse(dataItems[3]);
                        break;

                    case Ping.Right:
                        m_PingRight = int.Parse(dataItems[3]);
                        break;

                    default:
                        m_PingCenter = m_MaxDistancePing;
                        m_PingLeft = m_MaxDistancePing;
                        m_PingRight = m_MaxDistancePing;
                        break;
                }

                return true;
            }
            return false;
        }

        public bool UpdateStatus()
        {
            List<string> dataItems = GetTelemetry("S");
            if (dataItems != null && dataItems.Count == 7)
            {
                m_KinectPos = int.Parse(dataItems[0]);
                m_DrivePower = (int.Parse(dataItems[1]) == 1) ? true : false;
                m_Speed = int.Parse(dataItems[2]);
                m_Direction = (Direction)int.Parse(dataItems[3]);
                m_WheelPosRight = int.Parse(dataItems[4]);
                m_WheelPosLeft = int.Parse(dataItems[5]);
                m_HasArrived = (int.Parse(dataItems[6]) == 1) ? true : false;
                return true;
            }
            return false;
        }

        public void WaitArrived()
        {
            m_HasArrived = false;
            Thread.Sleep(500);

            while (!m_HasArrived)
            {
                UpdateStatus();
                Thread.Sleep(50);
            }
        }

        #region Properties

        public int IRLeft
        {
            get { return FuseReading(m_IRLeft, 0, m_MaxDistanceIR); }
        }

        public int IRCenter
        {
            get { return FuseReading(m_IRCenter, 0, m_MaxDistanceIR); }
        }

        public int IRRight
        {
            get { return FuseReading(m_IRRight, 0, m_MaxDistanceIR); }
        }

        public int PingLeft
        {
            get { return FuseReading(0, m_PingLeft, m_MaxDistancePing); }
        }

        public int PingCenter
        {
            get { return FuseReading(0, m_PingCenter, m_MaxDistancePing); }
        }

        public int PingRight
        {
            get { return FuseReading(0, m_PingRight, m_MaxDistancePing); }
        }

        public int KinectPosition
        {
            get { return m_KinectPos; }
        }

        public bool DrivePower
        {
            get { return m_DrivePower; }
        }

        public int CurrentSpeed
        {
            get { return m_Speed; }
        }

        public Direction CurrentDirection
        {
            get { return m_Direction; }
        }

        public int WheelPositionLeft
        {
            get { return m_WheelPosLeft; }
        }

        public int WheelPositionRight
        {
            get { return m_WheelPosRight; }
        }

        public bool HasArrived
        {
            get { return m_HasArrived; }
        }

        public int MaxDistanceIR
        {
            get { return m_MaxDistanceIR; }
            set { m_MaxDistanceIR = value; }
        }

        public int MaxDistancePing
        {
            get { return m_MaxDistancePing; }
            set { m_MaxDistancePing = value; }
        }

        #endregion

        #region Helpers

        private void SetComPort(string portName)
        {
            if (m_Port != null)
                m_Port.Close();

            m_Port = new SerialPort(portName, 19200);
            m_Port.DataBits = 8;
            m_Port.StopBits = StopBits.One;
            m_Port.Parity = Parity.None;
            m_Port.Handshake = Handshake.None;
            m_Port.WriteTimeout = 500;
            m_Port.ReadTimeout = 500;
            m_Port.NewLine = "\r";

            m_Port.Open();
        }

        private void SendCommand(string cmd)
        {
            if (m_Port.IsOpen)
            {
                string msg = ">RK" + cmd;
                m_Port.WriteLine(msg);
            }
        }

        private List<string> GetTelemetry(string frame)
        {
            SendCommand("T" + frame);

            string tel = "";
            try
            {
                tel = m_Port.ReadLine();
            }
            catch
            {
            }

            if (tel.StartsWith("<RK"))
            {
                string[] data = tel.Substring(3).Split(',');
                return data.ToList<string>();
            }

            return null;
        }

        private int FuseReading(int valueIR, int valuePing, int maxDistance)
        {
            if (valueIR > 0)
            {
                return Math.Min(valueIR, maxDistance);
            }
            else
            {
                if (valuePing > 0)
                    return Math.Min(valuePing, maxDistance);
                else
                    return maxDistance;
            }
        }

        #endregion
    }
}
